unit HugeModl; {Functions for `huge' (>64K) arrays under TPW v1.
                Copyright  1992, Jon Shemitz/Midnight Beach	}

{$define Huge_Inline}

{If "Huge_InLine" is $define-d, the pointer math routines become
 (relatively large) inline macros.}

{$C Moveable Discardable PreLoad}

interface

type	DGig = 0..$7FFFFFFF;	{Unsigned longs please, Borland!}

const	SegSize = $10000;	{64K}

var	SegInc: word;
	{This `magic' selector-tiling value is set by the init code}

{Alloc and free large blocks of memory}

  procedure GetHuge(var Result; Size: DGig);
  procedure FreeHuge(Huge: pointer);

  procedure GetMem(var Ptr; Size: DGig);
  procedure FreeMem(Ptr: pointer; Size: DGig);
  {These call System fns if Size < SegSize.  It would save some time
   to use HeapLimit, so that I pass requests GTE HeapLimit and LT
   SegSize directly to Windows (instead of through the System unit)
   but I *don't* like the possibilities for disaster if an app 
   changes HeapLimit in midstream ....}

{`Huge' pointer arithmatic}

  function OfsOf(Base: pointer; Ofs: DGig): pointer;

  {$ifdef Huge_InLine}		{Subroutines are commented}
  InLine($5E /			{pop    si}
         $5F /			{pop	di}
         $58 /			{pop	ax}
         $5A /			{pop	dx}
         $31/$C9 /		{xor    xc,cx}
         $01/$F0 /		{add    ax,si}
         $11/$F9 /		{adc    cx,di}
         $74/$06 /		{jz	@Tiled}
    {@Tile:}
    	 $03/$16/ SegInc / 	{add    dx,[SegInc]}
         $E2/$FA 		{loop   @Tile}		);
    {@Tiled}
  {$endif}

  function PreInc(var Pntr; Incr: DGig): pointer;
  {Modifies pointer Pntr and returns *new* value}

  {$ifdef Huge_InLine}
    InLine(
  	 $5B /			{pop    bx}
  	 $5E /			{pop    si}
         $5F /			{pop	di}
         $07 /			{pop    es}
         $26/$8B/$05 /		{mov    ax,[es:di]}
         $26/$8B/$55/$02 /	{mov    dx,[es:di+2]}
         $31/$C9 /		{xor    xc,cx}
         $01/$D8 /		{add    ax,bx}
         $11/$F1 /		{adc    cx,si}
         $74/$0A /		{jz	@Offset}
    {@Loop:}
    	 $03/$16/ SegInc / 	{add    dx,[SegInc]}
         $E2/$FA /		{loop   @Loop}
         $26/$89/$55/$02 /	{mov    [es:di+2],dx}
    {@Offset:}
         $26/$89/$05		{mov    [es:di],ax}	);
  {$endif}

  function PreIncW(var Pntr; Incr: word): pointer;
  {Modifies pointer Pntr and returns *new* value}

  {$ifdef Huge_InLine}

  InLine($5B /			{pop    bx}
	 $5F /			{pop	di}
         $07 /			{pop    es}
         $26/$8B/$05 /		{mov    ax,[es:di]}
         $26/$8B/$55/$02 /	{mov    dx,[es:di+2]}
	 $01/$D8 /		{add    ax,bx}
	 $73/$08 /		{jnc	@Offset}
	 $03/$16/ SegInc / 	{add    dx,[SegInc]}
	 $26/$89/$55/$02 /	{mov    [es:di+2],dx}
    {@Offset:}
         $26/$89/$05		{mov    [es:di],ax}	);
  {$endif}

  function PostInc(var Pntr; Incr: DGig): pointer;
  {Modifies pointer Pntr and returns *original* value}

  {$ifdef Huge_InLine}

  InLine($5B /			{pop    bx}
  	 $5E /			{pop    si}
         $5F /			{pop	di}
         $07 /			{pop    es}
         $26/$8B/$05 /		{mov    ax,[es:di]}
         $26/$8B/$55/$02 /	{mov    dx,[es:di+2]}
         $31/$C9 /		{xor    xc,cx}
         $01/$D8 /		{add    ax,bx}
         $11/$F1 /		{adc    cx,si}
         $74/$0A /		{jz	@Offset}
    {@Loop:}
    	 $03/$16/ SegInc / 	{add    dx,[SegInc]}
         $E2/$FA /		{loop   @Loop}
         $26/$87/$55/$02 /	{xchg   [es:di+2],dx}
    {@Offset:}
         $26/$87/$05		{xchg   [es:di],ax}	);
  {$endif}

  function PostIncW(var Pntr; Incr: word): pointer;
  {Modifies pointer Pntr and returns *original* value}
  {$ifdef Huge_InLine}

  InLine($5B /			{pop    bx}
	 $5F /			{pop	di}
         $07 /			{pop    es}
         $26/$8B/$05 /		{mov    ax,[es:di]}
         $26/$8B/$55/$02 /	{mov    dx,[es:di+2]}
	 $01/$D8 /		{add    ax,bx}
	 $73/$08 /		{jnc	@Offset}
	 $03/$16/ SegInc / 	{add    dx,[SegInc]}
	 $26/$87/$55/$02 /	{xchg   [es:di+2],dx}
    {@Offset:}
         $26/$87/$05		{xchg   [es:di],ax}	);
  {$endif}

  function PreDecW(var Pntr; Decrement: word): pointer;
  {Modifies pointer Pntr and returns *new* value}

{A "huge model" Move()}

  procedure Move(var Src, Dst; Bytes: DGig);
  {Note that this does NOT detect overlap, and always moves data
   in address order (Src first, Src[Bytes-1] last) }

  procedure Move3(var Src, Dst; Bytes: DGig);
  {Assumes that Src and Dst are ptrs:  Does
   Move(PostInc(Src, Bytes)^, PostInc(Dst, Bytes)^, Bytes}

{A "huge model" FillChar() and FillWord()}

  procedure FillChar(var Target; Chars: DGig; FillWith: char);

  procedure FillWord(var Target; Words: DGig; FillWith: word);

implementation

uses {$ifdef Windows} WinTypes, WinProcs {$else} WinAPI {$endif};

{Alloc and free large blocks of memory}

procedure GetHuge(var Result; Size: DGig);
var
  Handle:	tHandle;
  Ptr:		pointer absolute Result;
begin
  Handle := GlobalAlloc(gmem_Fixed, Size); {Ask Windows for memory}
  {as per TPW Programmer's Guide, p195}
  if Handle = 0
    then Ptr := Nil
    else Ptr := GlobalLock(Handle);	   {Get a pointer to it}
end;

type	DWord = record Lo, Hi: word end;
	{If this weren't for publication, I'd just use the
	 declaration in my GRABBAG unit.  Everyone should have one.}

procedure FreeHuge(Huge: pointer);
var
  Handle: LongInt;
begin
  Handle := GlobalHandle(DWord(Huge).Hi);
  {GlobalHandle() takes a segment and returns segment:handle}
  if DWord(Handle).Lo = 0
    then RunError(204)
    else GlobalFree(DWord(Handle).Lo);
    	 {Give memory back to Windows.  Note that there's no need to
	  call GlobalUnLock as protected mode Windows never changes
          lock count of non-discardable objects.}
end;

procedure GetMem(var Ptr; Size: DGig);
{Calls System.GetMem if Size < $10000}
var
  P: pointer absolute Ptr;
begin
  if Size <= $FFFF
    then System.GetMem(P, Size)
    else GetHuge(Ptr, Size);
end;

procedure FreeMem(Ptr: pointer; Size: DGig);
begin
  if Size <= $FFFF
    then System.FreeMem(Ptr, Size)
    else FreeHuge(Ptr);
end;

{`Huge' pointer arithmatic}

{$ifNdef Huge_InLine}

  function OfsOf(Base: pointer; Ofs: DGig): pointer; assembler;
  asm
	  mov	ax,[word ptr Base]
          mov	dx,[word ptr Base+2]	{dx:ax := Base}
          xor	cx,cx
          add	ax,[word ptr Ofs]	{new offset}
          adc	cx,[word ptr Ofs+2]	{# of 64K tiles to advance}
          jz	@Tiled
  @Tile:  add	dx,[SegInc]	{with very small CX's, this loop }
	  loop	@Tile		{is probably faster than MUL even}
  @Tiled:			{on a 486, and saves a couple of }
  end;				{mov instructions		 }

  function PreInc(var Pntr; Incr: DGig): pointer;	assembler;
  {Modifies pointer Ptr and returns *new* value}
  asm
	  les	di,[Pntr]		{pointer to the pointer}
          mov	ax,[es:di]
          mov	dx,[es:di+2]		{dx:ax := Ptr}
          xor	cx,cx
          add	ax,[word ptr Incr]	{new offset}
          adc	cx,[word ptr Incr+2]	{# of 64K tiles past Ptr}
          jz	@Offset
  @AddTile:
	  add	dx,[SegInc]
          loop	@AddTile
	  mov	[es:di+2],dx		{write back new selector}
  @Offset:mov 	[es:di],ax		{write back new offset}
  end;

  function PreIncW(var Pntr; Incr: word): pointer;	assembler;
  {Modifies pointer Ptr and returns *new* value}
  asm
	  les	di,[Pntr]		{pointer to the pointer}
          mov	ax,[es:di]
          mov	dx,[es:di+2]		{dx:ax := Ptr}
          add	ax,[word ptr Incr]	{new offset}
          jnc	@Offset
	  add	dx,[SegInc]
	  mov	[es:di+2],dx		{write back new selector}
  @Offset:mov 	[es:di],ax		{write back new offset}
  end;

  function PostInc(var Pntr; Incr: DGig): pointer;	assembler;
  {Modifies pointer Pntr and returns *original* value}
  asm
	  les	di,[Pntr]		{pointer to the pointer}
          mov	ax,[es:di]
          mov	dx,[es:di+2]		{dx,ax := Ptr}
          xor	cx,cx
          add	ax,[word ptr Incr]	{new offset}
          adc	cx,[word ptr Incr+2]	{# of 64K tiles past Ptr}
          jz	@Offset
  @AddTile:
	  add	dx,[SegInc]
          loop	@AddTile
	  xchg	[es:di+2],dx	{write new selector, get old}
  @Offset:xchg	[es:di],ax	{write new offset, get old}
  end;

  function PostIncW(var Pntr; Incr: word): pointer;	assembler;
  {Modifies pointer Pntr and returns *original* value}
  asm
	  les	di,[Pntr]		{pointer to the pointer}
          mov	ax,[es:di]
          mov	dx,[es:di+2]		{dx,ax := Ptr}
          add	ax,[Incr]		{new offset}
	  jnc	@Offset			{w/in segment?}
	  add	dx,[SegInc]		{no: tile the selector}
	  xchg	[es:di+2],dx	{write new selector, get old}
  @Offset:xchg	[es:di],ax	{write new offset, get old}
  end;

{$endif}

function PreDecW(var Pntr; Decrement: word): pointer; assembler;
{Modifies pointer Pntr and returns *new* value}
asm
	  les	di,[Pntr]		{pointer to the pointer}
          mov	ax,[es:di]
          mov	dx,[es:di+2]		{dx:ax := Ptr}
          sub	ax,[word ptr Decrement]	{new offset}
          jnc	@Offset
	  sub	dx,[SegInc]
	  mov	[es:di+2],dx		{write back new selector}
  @Offset:mov 	[es:di],ax		{write back new offset}
end;

{A "huge model" Move()}

type
  Segment = 0..SegSize {bytes} ;

  procedure WordMove(var Src, Dst; Bytes: Segment); near; assembler;
  asm
  	mov	dx,ds
        cld
	lds	si,[Src]
        les	di,[Dst]
        mov	cx,[word ptr Bytes]
        rcr	[byte ptr Bytes+2],1	{cf := low bit}
        rcr	cx,1			{Cnvrt bytes to words}
	rep	movsw
	adc	cx,0
	rep	movsb			{Move odd byte, if any}
        mov	ds,dx
  end;

procedure Move(var Src, Dst; Bytes: DGig);
var
  Chunk:          Segment;
  MaxOfs:	  word ;
  SrcPtr, DstPtr: pointer;
begin
  SrcPtr := @ Src;
  DstPtr := @ Dst;
  while Bytes > 0 do
    begin
    asm				{MaxOfs := Max(Ofs(Src), Ofs(Dst) }
    	mov	ax,[word ptr SrcPtr]
        cmp	ax,[word ptr DstPtr]
        jae	@Src
        mov	ax,[word ptr DstPtr]
      @Src:
      	mov	[MaxOfs],ax
    end;
    Chunk := SegSize - MaxOfs;
    {# of bytes that can be moved w/out crossing a segment boundary}
    if Chunk > Bytes then Chunk := Bytes;
    WordMove( PostInc(SrcPtr, Chunk)^,
    	      PostInc(DstPtr, Chunk)^,
	      Chunk );
    Dec(Bytes, Chunk);
    end;
end;

procedure Move3(var Src, Dst; Bytes: DGig);
{Assumes that Src and Dst are ptrs}
begin
  Move(PostInc(Src, Bytes)^, PostInc(Dst, Bytes)^, Bytes);
end;

{A "huge model" FillChar() and FillWord()}

  procedure FillB(var Target; Count: Segment; FillWith: char); assembler; {In-segment fill}
  asm
	les	di,[Target]
	mov	al,[FillWith]
        mov	ah,al				{Replicate the fill char}
	mov	cx,[word ptr Count]             {Low word}
	rcr	[word ptr Count + 2], 1		{High word}
        rcr	cx,1				{Count div 2; remainder in CF}
	cld
	rep	stosw
	adc	cx,0
        rep 	stosb
  end;

procedure FillChar(var Target; Chars: DGig; FillWith: char);
var
  Chunk: Segment;
  Dst:   pointer;
  Count: LongInt absolute Chars;
begin
  Dst := @ Target;
  while Count > 0 do
    begin
{*} Chunk := SegSize - Ofs(Dst^);
    Dec(Count, Chunk);
    if Count < 0 then Inc(Chunk, Count);
    FillB(PostInc(Dst, Chunk)^, Chunk, FillWith);
    end;
end;

  procedure FillW(var Target; Count: Segment; FillWith: word); assembler; {In-segment fill}
  asm
	les	di,[Target]
	mov	ax,[FillWith]
	mov	cx,[word ptr Count]             {Low word}
	cld
	rep	stosw
  end;

procedure FillWord(var Target; Words: DGig; FillWith: word);
var
  Chunk: Segment;
  Dst:   pointer;
  Count: LongInt absolute Words;
begin
  Dst := @ Target;
  {$ifopt R+}
    if Odd(Ofs(Dst^)) then RunError(201);
  {$endif}
  while Count > 0 do
    begin
{*} Chunk := (SegSize - Ofs(Dst^)) shr 1;
    Dec(Count, Chunk);
    if Count < 0 then Inc(Chunk, Count);
    FillW(PostInc(Dst, Chunk shl 1)^, Chunk, FillWith);
    end;
end;


{Init code}

{$ifdef Windows}
  procedure __ahincr; far; external 'kernel';
{$endif}

{A little black magic underneath it all, just to
 keep things interesting.  You've gotta love Microsoft.}

begin
  SegInc := {$ifdef Windows} Ofs(__ahincr);
            {$else}	     SelectorInc;      	{$endif}
end.